from typing import ClassVar

import attr
import numpy as np

from arlunio.math import normalise


@attr.s(auto_attribs=True, repr=False)
class Rays:
    """A collection of rays"""

    origin: np.ndarray
    """The point where the rays originate from."""

    direction: np.ndarray
    """An array of vectors indicating which direction each ray is pointing"""

    def __repr__(self):
        return f"Rays(origin={self.origin.shape}, direction={self.direction.shape})"

    def __getitem__(self, key):
        return Rays(self.origin[key], self.direction[key])

    def at(self, t):
        """Get the position for each value of t."""
        return self.origin + t[:, np.newaxis] * self.direction


@attr.s(auto_attribs=True, repr=False)
class ScatterPoint:
    """A scatter point is where rays have intersected an object and where various
    calculations should take place."""

    MAX_FLOAT: ClassVar[float] = np.finfo(np.float64).max

    hit: np.ndarray
    """A boolean array with shape :code:`(n,)` indicating which rays have intersected
    some object."""

    t: np.ndarray
    """A scalar array with shape :code:`(n,)` indicating the value of the parameter
    :code:`t` at the point of intersection."""

    p: np.ndarray
    """An array of vectors with shape :code:`(n, 3)` indicating the point of
    intersection in space."""

    normal: np.ndarray
    """An array of vectors with shape :code:`(n, 3)` indicating the direction of the
    surface normal at the point of intersection."""

    front_face: np.ndarray
    """A boolean array with shape :code:`(n, 3)` indicating which side of the surface
    the incoming ray was on."""

    def __repr__(self):

        props = {
            "hit": self.hit.shape,
            "t": self.t.shape,
            "p": self.p.shape,
            "normal": self.normal.shape,
            "front_face": self.front_face.shape,
        }

        args = ", ".join([f"{k}={v}" for k, v in props.items()])
        return f"ScatterPoint({args})"

    def __getitem__(self, key):

        hit = self.hit[key]
        t = self.t[key]
        p = self.p[key]
        normal = self.normal[key]
        front_face = self.front_face[key]

        return ScatterPoint(hit, t, p, normal, front_face)

    def merge(self, other):
        """Merge the results of another scatter point in with this one."""

        self.hit = np.logical_or(self.hit, other.hit)

        self.t[other.hit] = other.t
        self.p[other.hit] = other.p
        self.normal[other.hit] = other.normal
        self.front_face[other.hit] = other.front_face

    @classmethod
    def new(cls, rays: Rays):
        """Create a new scatter point based on an initial cluster of rays."""

        n, _ = rays.direction.shape

        true = np.full((n,), True)
        false = np.full((n,), False)
        maxf = np.full((n,), cls.MAX_FLOAT)

        params = {
            "hit": false,
            "t": maxf,
            "p": rays.direction,
            "normal": normalise(rays.direction),
            "front_face": true,
        }

        return cls(**params)



import numpy as np
import numpy.random as npr

import arlunio as ar
from .data import Rays
from .data import ScatterPoint
from arlunio.color import getcolorf


def random_unit_sphere(n):
    a = npr.rand(n) * 2 * np.pi
    z = (npr.rand(n) * 2) - 1
    r = np.sqrt(1 - (z * z))

    return np.dstack([r * np.cos(a), r * np.sin(a), z])[0]


@ar.definition
def Gradient(scatter: ScatterPoint, *, color1=None, color2=None):
    """A basic gradient color."""

    color1 = np.array([1.0, 1.0, 1.0]) if color1 is None else color1
    color2 = np.array([0.5, 0.7, 1.0]) if color2 is None else color2

    t = 0.5 * (scatter.normal[:, 1]) + 1

    start = np.einsum("n,np->np", (1 - t), color1.reshape(1, 3))
    stop = np.einsum("n,np->np", t, color2.reshape(1, 3))

    return start + stop


@ar.definition
def NormalMap(scatter: ScatterPoint):
    """Visualise the normals on a given surface."""

    r = scatter.normal[:, 0] + 1
    g = scatter.normal[:, 1] + 1
    b = scatter.normal[:, 2] + 1

    return 0.5 * np.dstack([r, g, b])[0]


@ar.definition
def LambertianDiffuse(scatter: ScatterPoint, *, color="lightgrey"):
    """Lambertian diffuse material."""

    color = getcolorf(color)
    n = scatter.p.shape[0]

    origin = scatter.p
    directions = scatter.normal + random_unit_sphere(n)

    return (color, Rays(origin, directions))



import PIL.ImageColor as Color


def getcolor(*args, **kwargs):
    """Exactly as Pillow's getrgb function."""

    return Color.getcolor(*args, **kwargs)


def getcolorf(*args, **kwargs):
    """Exactly as Pillow's getrgb function, but the values are returned as floats."""

    col = Color.getcolor(*args, **kwargs)
    return tuple([c / 255 for c in col])



from typing import Callable

import numpy as np

import arlunio as ar


def clamp(vs, min_=0, max_=1):
    """Force an array of values to stay within a range of values.

    Parameters
    ----------
    vs:
      The array of values to clamp
    min_:
      The minimum value the result should contain
    max_:
      The maximum value the result should contain

    Examples
    --------

    By default values will be limited to between :code:`0` and :code:`1`

    >>> from arlunio.math import clamp
    >>> import numpy as np
    >>> vs = np.linspace(-1, 2, 6)
    >>> clamp(vs)
    array([0. , 0. , 0.2, 0.8, 1. , 1. ])

    But this can be changed with extra arguments to the :code:`clamp` function

    >>> clamp(vs, min_=-1, max_=0.5)
    array([-1. , -0.4,  0.2,  0.5,  0.5,  0.5])
    """
    vs = np.array(vs)
    vs[vs > max_] = max_
    vs[vs < min_] = min_

    return vs


def dot(us, vs):
    """Return the dot product of two arrays of vectors."""
    return np.sum(us * vs, axis=-1)


def length(vs):
    return np.sqrt(np.sum(vs * vs, axis=-1))


def lerp(start: float = 0, stop: float = 1) -> Callable[[float], float]:
    """Return a function that will linerarly interpolate between a and b.

    Parameters
    ----------
    start:
      The value the interpolation should start from.
    stop:
      The value the interpolation should stop at.

    Examples
    --------

    By default this function will interpolate between :code:`0` and :code:`1`

    >>> from arlunio.math import lerp
    >>> f = lerp()
    >>> f(0)
    0
    >>> f(1)
    1

    However by passing arguments to the :code:`lerp` function we can change the bounds
    of the interpolation.

    >>> import numpy as np
    >>> ts = np.linspace(0, 1, 4)
    >>> f = lerp(start=3, stop=-1)
    >>> f(ts)
    array([ 3.        ,  1.66666667,  0.33333333, -1.        ])

    """

    def f(t: float) -> float:
        return (1 - t) * start + t * stop

    return f


def normalise(vs):
    """Normalise an array into the range :math:`[0, 1]`

    Parameters
    ----------
    vs:
       The array to normalise.
    """

    if len(vs.shape) == 1:
        return vs / length(vs)

    if len(vs.shape) == 2:
        return vs / length(vs)[:, np.newaxis]


@ar.definition
def X(width: int, height: int, *, x0=0, scale=1, stretch=False):
    """Cartesian :math:`x` coordinates.

    .. arlunio-image:: X Coordinates
       :align: right

       ::

          from arlunio.math import X
          from arlunio.image import colorramp

          x = X()
          image = colorramp(x(width=256, height=256))

    Attributes
    ----------
    x0:
        Shift all the coordinate values by :code:`x0`
    scale:
        Controls the magnitude of the extreme values.
    stretch:
        If :code:`True` and the image is wider than it is tall then the grid will be
        stretched so that :code:`x = scale` falls on the image border. Otherwise the
        image's aspect ratio will be taken into account and :code:`x = scale` will fall
        somewhere within the boundaries of the image.

    Examples
    --------

    By default values will be generated between :math:`\\pm 1`::

       >>> from arlunio.math import X
       >>> x = X()
       >>> x(width=4, height=4)
       array([[-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ]])

    If however the image is wider than it is tall this range will be extended so that
    the resulting image is not stretched::

       >>> x(width=4, height=2)
       array([[-2.        , -0.66666667,  0.66666667,  2.        ],
              [-2.        , -0.66666667,  0.66666667,  2.        ]])

    This behaviour can be disabled with the :code:`stretch` attribute::

       >>> x.stretch = True
       >>> x(width=4, height=2)
       array([[-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ]])

    Additionally the :code:`scale` attribute can be used to adjust the magnitude of the
    extreme values generated, while the :code:`x0` attribute can be used to shift all
    the values by a given amount::

       >>> x = X(x0=-2, scale=2)
       >>> x(width=4, height=4)
       array([[0.        , 1.33333333, 2.66666667, 4.        ],
              [0.        , 1.33333333, 2.66666667, 4.        ],
              [0.        , 1.33333333, 2.66666667, 4.        ],
              [0.        , 1.33333333, 2.66666667, 4.        ]])

    """
    ratio = width / height

    if not stretch and ratio > 1:
        scale = scale * ratio

    x = np.linspace(-scale, scale, width)
    x = np.array([x for _ in range(height)])

    return x - x0


@ar.definition
def Y(width: int, height: int, *, y0=0, scale=1, stretch=False):
    """Cartesian :math:`y` coordinates

    .. arlunio-image:: Y Coordinates
       :align: right

       ::

          from arlunio.math import Y
          from arlunio.image import colorramp

          y = Y()
          image = colorramp(y(width=256, height=256))

    Attributes
    ----------
    y0:
        Shift all the coordinate values by :code:`y0`
    scale:
        Controls the size of the extreme values
    stretch:
        If :code:`True` and the image is taller than it is wide then the grid will be
        stretched so that :code:`y = scale` falls on the border. Otherwise the image's
        aspect ratio will be taken into account and :code:`y = scale` will fall
        somewhere within the boundaries of the image.

    Examples
    --------

    By default values will be generated between :math:`\\pm 1`::

       >>> from arlunio.math import Y
       >>> y = Y()
       >>> y(width=4, height=4)
       array([[ 1.        ,  1.        ,  1.        ,  1.        ],
              [ 0.33333333,  0.33333333,  0.33333333,  0.33333333],
              [-0.33333333, -0.33333333, -0.33333333, -0.33333333],
              [-1.        , -1.        , -1.        , -1.        ]])

    If however the image is taller than it is wide this range will be extended so that
    the resulting image is not stretched::

       >>> y(width=2, height=4)
       array([[ 2.        ,  2.        ],
              [ 0.66666667,  0.66666667],
              [-0.66666667, -0.66666667],
              [-2.        , -2.        ]])

    This behaviour can be disabled with the :code:`stretch` attribute::

       >>> y.stretch = True
       >>> y(width=2, height=4)
       array([[ 1.        ,  1.        ],
              [ 0.33333333,  0.33333333],
              [-0.33333333, -0.33333333],
              [-1.        , -1.        ]])

    Additionally the :code:`scale` attribute can be used to adjust the magnitude of the
    extreme values generated, while the :code:`y0` attribute can be used to shift all
    the values by a given amount::

       >>> y = Y(y0=-2, scale=2)
       >>> y(width=4, height=4)
       array([[4.        , 4.        , 4.        , 4.        ],
              [2.66666667, 2.66666667, 2.66666667, 2.66666667],
              [1.33333333, 1.33333333, 1.33333333, 1.33333333],
              [0.        , 0.        , 0.        , 0.        ]])

    """
    ratio = height / width

    if not stretch and ratio > 1:
        scale = scale * ratio

    y = np.linspace(scale, -scale, height)
    y = np.array([y for _ in range(width)]).transpose()

    return y - y0


@ar.definition
def R(x: X, y: Y):
    """Polar :math:`r` coordinates.

    .. arlunio-image:: R Coordinates
       :align: right

       ::

          from arlunio.math import R
          from arlunio.image import colorramp

          r = R()
          image = colorramp(r(width=256, height=256))

    This definition corresponds with the distance a given point is from the origin and
    can be calculated from the point's equivalent Cartesian coordinate representation

    .. math::

       r = \\sqrt{x^2 + y^2}

    Examples
    --------
    ::

       >>> from arlunio.math import R
       >>> r = R()
       >>> r(width=5, height=5)
       array([[1.41421356, 1.11803399, 1.        , 1.11803399, 1.41421356],
              [1.11803399, 0.70710678, 0.5       , 0.70710678, 1.11803399],
              [1.        , 0.5       , 0.        , 0.5       , 1.        ],
              [1.11803399, 0.70710678, 0.5       , 0.70710678, 1.11803399],
              [1.41421356, 1.11803399, 1.        , 1.11803399, 1.41421356]])

    While this definition does not currently have any attributes of its own, since it's
    derived from the :class:`arlunio.math.X` and :class:`arlunio.math.Y` definitions it
    automatically inherits the attributes from these base definitions::

       >>> r = R(x0=-2, y0=-2, scale=2)
       >>> r(width=5, height=5)
       array([[4.        , 4.12310563, 4.47213595, 5.        , 5.65685425],
              [3.        , 3.16227766, 3.60555128, 4.24264069, 5.        ],
              [2.        , 2.23606798, 2.82842712, 3.60555128, 4.47213595],
              [1.        , 1.41421356, 2.23606798, 3.16227766, 4.12310563],
              [0.        , 1.        , 2.        , 3.        , 4.        ]])

    Notice how in the case where the base definitions share a attribute (:code:`scale`
    in this case) they both share the value that is set when creating an instance of the
    :code:`R` definition.

    """
    return np.sqrt(x * x + y * y)


@ar.definition
def T(x: X, y: Y, *, t0=0):
    """Polar :math:`\\theta` coordinates.

    .. arlunio-image:: T Coordinates
       :align: right

       ::

          from arlunio.math import T
          from arlunio.image import colorramp

          t = T()
          image = colorramp(t(width=256, height=256))

    This definition corresponds with the angle a given point is around from the positive
    :math:`x`-axis. This can be calculated from the point's equivalent Cartesian
    coordinate representation. All angles are given in :term:`radians`

    .. math::

       t = atan2\\left(\\frac{y}{x}\\right)

    Attributes
    ----------
    t0:
        Shift all the coordinate values by :code:`t0`

    Examples
    --------

    By default all point on the :math:`x`-axis will have a value of :code:`t = 0`::

       >>> from arlunio.math import T
       >>> t = T()
       >>> t(width=5, height=5)
       array([[ 2.35619449,  2.03444394,  1.57079633,  1.10714872,  0.78539816],
              [ 2.67794504,  2.35619449,  1.57079633,  0.78539816,  0.46364761],
              [ 3.14159265,  3.14159265,  0.        ,  0.        ,  0.        ],
              [-2.67794504, -2.35619449, -1.57079633, -0.78539816, -0.46364761],
              [-2.35619449, -2.03444394, -1.57079633, -1.10714872, -0.78539816]])

    This can be changed however with the :code:`t0` attribute::

       >>> from math import pi
       >>> t.t0 = pi
       >>> t(width=5, height=5)
       array([[-0.78539816, -1.10714872, -1.57079633, -2.03444394, -2.35619449],
              [-0.46364761, -0.78539816, -1.57079633, -2.35619449, -2.67794504],
              [ 0.        ,  0.        , -3.14159265, -3.14159265, -3.14159265],
              [-5.8195377 , -5.49778714, -4.71238898, -3.92699082, -3.60524026],
              [-5.49778714, -5.17603659, -4.71238898, -4.24874137, -3.92699082]])

    Also, being a definition derived from :class:`arlunio.math.X` and
    :class:`arlunio.math.Y` the attributes for these definitions are also available to
    control the output::

       >>> t = T(x0=-2, y0=-2, scale=2)
       >>> t(width=5, height=5)
       array([[1.57079633, 1.32581766, 1.10714872, 0.92729522, 0.78539816],
              [1.57079633, 1.24904577, 0.98279372, 0.78539816, 0.64350111],
              [1.57079633, 1.10714872, 0.78539816, 0.5880026 , 0.46364761],
              [1.57079633, 0.78539816, 0.46364761, 0.32175055, 0.24497866],
              [0.        , 0.        , 0.        , 0.        , 0.        ]])

    """
    t = np.arctan2(y, x)
    return t - t0


@ar.definition
def Cartesian(x: X, y: Y):
    """Cartesian coordinates."""
    return np.dstack([x, y])


@ar.definition
def Polar(r: R, t: T):
    """Polar coordinates."""
    return np.dstack([r, t])


@ar.definition
def Barycentric(x: X, y: Y, *, a=(0.5, -0.5), b=(0, 0.5), c=(-0.5, -0.5)):
    """Barycentric coordinates.

    .. arlunio-image:: Barycentric Demo
       :align: right

       ::

          import arlunio.math as math
          import arlunio.image as img
          import numpy as np

          coords = math.Barycentric()(width=256, height=256)

          l1 = np.abs(coords[:, :, 0])
          l2 = np.abs(coords[:, :, 1])
          l3 = np.abs(coords[:, :, 2])

          r = img.colorramp(l1, start='red', stop='black')
          g = img.colorramp(l2, start='green', stop='black')
          b = img.colorramp(l3, start='blue', stop='black')

          rs = np.asarray(r)
          gs = np.asarray(g)
          bs = np.asarray(b)

          image = img.fromarray(rs + gs + bs)

    Returns the Barycentric coordinate grid as defined by the given triangle.

    Attributes
    ----------
    a:
      The cartesian :math:`(x, y)` coordinates of the point :math:`a`
    b:
      The cartesian :math:`(x, y)` coordinates of the point :math:`b`
    c:
      The cartesian :math:`(x, y)` coordinates of the point :math:`c`
    """

    h, w = x.shape

    # Construct the transform matrix T
    t1 = a[0] - c[0]
    t2 = b[0] - c[0]
    t3 = a[1] - c[1]
    t4 = b[1] - c[1]

    # Compute the inverse determinant
    d = 1 / ((t1 * t4) - (t2 * t3))
    p = np.dstack([x, y]).reshape(w * h, 2) - c

    p1 = p[:, 0]
    p2 = p[:, 1]

    # Compute the coordinates
    l1 = d * ((p1 * t4) - (p2 * t2))
    l2 = d * ((p2 * t1) - (p1 * t3))
    l3 = 1 - l1 - l2

    return np.dstack([l1, l2, l3])[0].reshape(h, w, 3)



import pathlib
import textwrap
import traceback
from typing import List

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.images import Figure
from docutils.statemachine import StringList
from sphinx.util import logging
from sphinx.util.nodes import nested_parse_with_titles

import arlunio.image as image


class Count:
    def __init__(self):
        self.i = 0

    def __call__(self):
        self.i += 1
        return self.i


class arlunio_image(nodes.General, nodes.Element):
    pass


def visit_arlunio_image(self, node):
    pass


def depart_arlunio_image(self, node):
    pass


def parse_content(state, content: StringList) -> List[nodes.Node]:
    """Given a reStructured text representation of some content format parse it
    into a list of nodes."""

    section = nodes.section()
    section.document = state.document
    nested_parse_with_titles(state, content, section)

    return section.children


def reformat_content(caption: str, code: str, include_code: bool) -> StringList:
    """Reformat the content of the arlunio-image directive so that it's compatible with
    the standard figure directive."""

    count = Count()
    src = "<arlunio_image>"

    content = StringList()
    content.append(caption, src, count())
    content.append("", src, count())

    if include_code:

        content.append(".. code-block:: python", src, count())
        content.append("", src, count())

        for line in code.splitlines():
            line = textwrap.indent(line, "   ")
            content.append(line, src, count())

    return content


def find_caption(doctree: List[nodes.Node]) -> str:
    """Given the doctree representing the content of the arlunio-image directive
    find and return the text representing the caption (if any)."""

    logger = logging.getLogger(__name__)
    text = ".."

    if len(doctree) > 0 and isinstance(doctree[0], nodes.paragraph):
        text = doctree[0].astext()
        logger.debug("[arlunio-image]: Found caption text '%s'", text)

    return text


def find_code(doctree: List[nodes.Node]) -> str:
    """Given the doctree representing the content of the arlunio-image directive
    find and return the source code for the image (if any)."""

    logger = logging.getLogger(__name__)
    logger.debug("[arlunio-image] doctree: %s", doctree)
    candidates = [n for n in doctree if isinstance(n, nodes.literal_block)]

    if len(candidates) == 0:
        raise ValueError("Missing image code.")

    if len(candidates) > 1:
        raise ValueError("Multiple code blocks is not supported.")

    code = candidates[0].astext()
    logger.debug("[arlunio-image]: Found source code: %s...", code[: code.find("\n")])

    return code


def render_image(src: str, name: str, outdir: pathlib.Path) -> pathlib.Path:
    """Given the source code for an image, render it and save it to file.

    Parameters
    ----------
    src:
        The code that produces the image
    name:
        The name of the image
    outdir:
        Where the image should be saved.

    Return
    ------
    The path to the image, relative to the root of the Sphinx project
    """
    env = {}
    logger = logging.getLogger(__name__)

    code = compile(src, "<string>", "exec")
    exec(code, env)

    img = None

    for obj in env.values():
        if isinstance(obj, image.Image):
            img = obj

    if img is None:
        raise ValueError("Code did not produce a valid image.")

    outfile = pathlib.Path("_images", name + ".png")
    logger.debug("[arlunio-image] Saving image to: %s", outfile)

    image.save(img, pathlib.Path(outdir, outfile), mkdirs=True)

    return outfile


class ArlunioImageDirective(Figure):
    """Given some code that produces an image, render it in the page."""

    has_content = True
    final_argument_whitespace = True

    option_spec = Figure.option_spec.copy()
    option_spec["gallery"] = directives.unchanged
    option_spec["include-code"] = directives.flag

    def run(self):

        logger = logging.getLogger(__name__)

        imgname = self.arguments[0]
        imgpath = imgname.lower().replace(" ", "-")

        logger.debug("[arlunio-image]: Rendering: %s", imgname)
        include_code = "include-code" in self.options.keys()

        # First we will process the content of the directive in order to produce an
        # image on disk. We will then defer to the default behavior of the Figure
        # directive to handle the rest.
        app = self.state.document.settings.env.app

        doctree = parse_content(self.state, self.content)
        caption = find_caption(doctree)

        try:
            src = find_code(doctree)
            imguri = render_image(src, imgpath, app.confdir)
        except Exception:
            tback = traceback.format_exc()

            # Flag the issue to the user to the issue in the log
            location = self.state_machine.get_source_and_line(self.lineno)
            logger.warning("Unable to render image\n%s", tback, location=location)

            # But also make the error obvious in the docs.
            message = nodes.Text("Unable to render image:")
            err = nodes.literal_block("", tback)

            return [nodes.error("", message, err)]

        # Rewrite the arguments list to contain the reference to the image file
        self.arguments[0] = "/" + str(imguri)

        # Rewrite the content according to the given options.
        content = reformat_content(caption, src, include_code)

        self.content_offset = 0
        self.content = content

        # Now defer to the standard figure implementation.
        (figure,) = Figure.run(self)
        logger.debug("[arlunio-image]: Figure: %s", figure)

        opts = {k: v for k, v in self.options.items() if k in ("gallery",)}
        opts["source"] = src
        opts["name"] = imgname
        opts["caption"] = caption

        return [arlunio_image("", figure, **opts)]


def register(app):

    app.add_node(
        arlunio_image,
        html=(visit_arlunio_image, depart_arlunio_image),
        latex=(visit_arlunio_image, depart_arlunio_image),
        text=(visit_arlunio_image, depart_arlunio_image),
    )

    app.add_directive("arlunio-image", ArlunioImageDirective)



import base64
import io
import logging
import pathlib
from typing import Optional

import numpy as np
import PIL.Image as PImage

import arlunio.color as color
import arlunio.mask as mask
import arlunio.math as math

logger = logging.getLogger(__name__)


class Image:
    """Our representation of an image, implemented as a wrapper around a standard
    Pillow image."""

    def __init__(self, img: PImage.Image):
        self.img = img
        """The wrapped pillow image object."""

    def __eq__(self, other):

        if not isinstance(other, Image):
            return False

        a = np.asarray(self.img)
        b = np.asarray(other.img)

        return (a == b).all()

    def __add__(self, other):

        if isinstance(other, Image):
            other = other.img

        if not isinstance(other, PImage.Image):
            raise TypeError("Addition is only supported between images.")

        img = self.copy()
        img.alpha_composite(other)

        return img

    @property
    def __array_interface__(self):
        # Ensure that our version of an image also plays nice with numpy.
        return self.img.__array_interface__

    def _repr_png_(self):
        # Give nice previews in jupyter notebooks
        return self.img._repr_png_()

    def alpha_composite(self, im, *args, **kwargs):
        """Composites an image onto this image.

        See :meth:`pillow:PIL.Image.Image.alpha_composite`
        """

        if isinstance(im, Image):
            im = im.img

        self.img.alpha_composite(im, *args, **kwargs)

    def copy(self):
        """Return a copy of the image.
        See :meth:`pillow:PIL.Image.Image.copy`
        """
        return Image(self.img.copy())

    def paste(self, *args, **kwargs):
        """Paste another image into this image.

        See :meth:`pillow:PIL.Image.Image.paste`
        """
        self.img.paste(*args, **kwargs)

    def save(self, *args, **kwargs):
        """Save the image with the given filename.

        See :meth:`pillow:PIL.Image.Image.save`
        """
        self.img.save(*args, **kwargs)


def new(*args, **kwargs):
    """Creates a new image with the given mode and size

    See :func:`pillow:PIL.Image.new`
    """
    return Image(PImage.new(*args, **kwargs))


def fromarray(*args, **kwargs):
    """Create an image from an array

    See :func:`pillow:PIL.Image.fromarray`
    """
    return Image(PImage.fromarray(*args, **kwargs))


def load(*args, **kwargs) -> Image:
    """Load an image from the given file.

    See :func:`pillow:PIL.Image.open`
    """
    return Image(PImage.open(*args, **kwargs))


def save(image: Image, filename: str, mkdirs: bool = False) -> None:
    """Save an image in PNG format.

    :param filename: The filepath to save the image to.
    :param mkdirs: If true, make any parent directories
    """
    path = pathlib.Path(filename)

    if not path.parent.exists() and mkdirs:
        path.parent.mkdir(parents=True)

    with open(filename, "wb") as f:
        image.save(f)


def encode(image: Image) -> bytes:
    """Return the image encoded as a base64 string.

    Parameters
    ----------
    image:
       The image to encode.

    Example
    -------
    ::

       >>> import arlunio.image as image
       >>> img = image.new("RGBA", (8, 8), color='red')
       >>> image.encode(img)
       b'iVBORw0KGgoAAAANSUhEUgAAAAgAAAAICAYAAADED76LAAAAFklEQVR4nGP8z8DwnwEPYMInOXwUAAASWwIOH0pJXQAAAABJRU5ErkJggg=='

    """

    with io.BytesIO() as byte_stream:
        image.save(byte_stream, "PNG")
        image_bytes = byte_stream.getvalue()

        return base64.b64encode(image_bytes)


def decode(bytestring: bytes) -> Image:
    """Decode the image represented by the given bytestring into an image object.

    Parameters
    ----------
    bytestring:
       The bytestring to decode.

    Example
    -------

    .. arlunio-image:: Decode Example
       :include-code:

       ::

          import arlunio.image as image

          bytestring = b'iVBORw0KGgoAAAANSUhEUgAAAAgAAAAICAYAAADED76LAAAAFklEQVR4nGP8z8DwnwEPYMInOXwUAAASWwIOH0pJXQAAAABJRU5ErkJggg=='  # noqa: E501
          img = image.decode(bytestring)
    """

    data = base64.b64decode(bytestring)
    bytes_ = io.BytesIO(data)

    return Image(load(bytes_))


def colorramp(values, start: Optional[str] = None, stop: Optional[str] = None) -> Image:
    """Given a 2d array of values, produce an image gradient based on them.

    .. arlunio-image:: Colorramp Demo
       :align: right

       ::

          import arlunio.image as image
          import arlunio.math as math
          import numpy as np

          cartesian = math.Cartesian()
          p = cartesian(width=256, height=256)
          x, y = p[:, :, 0], p[:, :, 1]

          values = np.sin(2*x*np.pi) * np.sin(2*y* np.pi)
          img = image.colorramp(values)

    First this function will scale the input array so that all values fall in the range
    :math:`[0, 1]`. It will then produce an image with the same dimensions as the
    original array. The color of each pixel will be chosen based on the corresponding
    value of the scaled array.

    - If the value is :math:`0` the color will be given by the :code:`start` parameter

    - If the value is :math:`1` the color will be given by the :code:`stop` parameter

    - Otherwise the color will be some mix between the two.

    Parameters
    ----------
    values:
       The array of values used to decide on the color.
    start:
       The color to use for values near :math:`0` (default, :code:`black`)
    stop:
       The color to use for values near :math:`1` (default, :code:`white`)

    Examples
    --------

    .. arlunio-image:: Colorramp Demo 2
       :include-code:

       ::

          import arlunio.image as image
          import arlunio.math as math
          import numpy as np

          cartesian = math.Cartesian()
          p = cartesian(width=256, height=256)

          bg = image.new("RGBA", (256, 256), color="black")
          x = image.colorramp(p[:, :, 0], start="#0000", stop="#f007")
          y = image.colorramp(p[:, :, 1], start="#0000", stop="#00f7")

          img = x + y
    """

    # Scale all the values so that they fall into the range [0, 1]
    minx = np.min(values)
    vs = np.array(values) - minx
    vs = vs / np.max(vs)

    if start is None:
        start = "black"

    if stop is None:
        stop = "white"

    start = color.getcolor(start, "RGBA")
    stop = color.getcolor(stop, "RGBA")

    funcs = [math.lerp(a, b) for a, b in zip(start, stop)]
    channels = [np.floor(func(vs)) for func in funcs]

    pixels = np.array(np.dstack(channels), dtype=np.uint8)
    return fromarray(pixels)


def fill(
    mask: mask.Mask,
    foreground: Optional[str] = None,
    background: Optional[str] = None,
    image: Optional[Image] = None,
) -> Image:
    """Apply color to an image, as specified by a mask.

    Parameters
    ----------
    mask:
        The mask that selects the region to be coloured
    foreground:
        A string representation of the color to use, this can be in any format that is
        supported by the :mod:`pillow:PIL.ImageColor` module. If omitted this will
        default to black.
    background:
        In the case where an existing image is not provided this parameter can be used
        to set the background color of the generated image. This can be any string that
        is accepted by the :mod:`pillow:PIL.ImageColor` module. If omitted this will
        default to transparent
    image:
        The image to color in, if omitted a blank image will be used.

    Example
    --------

    .. arlunio-image:: Fill Demo
       :include-code:

       ::

          import arlunio.image as image
          import arlunio.shape as shape

          circle = shape.Circle(x0=-0.5, y0=0.25, r=0.6)
          img = image.fill(circle(width=512, height=256), foreground='red')

          circle.x0, circle.y0 = 0, 0
          img = image.fill(circle(width=512, height=256), foreground='#0f0', image=img)

          circle.x0, circle.y0 = 0.5, -0.25
          img = image.fill(circle(width=512, height=256), foreground='blue', image=img)

    """

    foreground = "#000" if foreground is None else foreground
    fill_color = color.getcolor(foreground, "RGBA")

    mask_img = PImage.fromarray(mask)

    if image is None:
        background = "#0000" if background is None else background

        height, width = mask.shape
        image = new("RGBA", (width, height), color=background)

    else:
        image = image.copy()

    image.paste(fill_color, mask=mask_img)

    return image



import functools
from typing import Union

import numpy as np

import arlunio as ar


class Mask(np.ndarray):
    """A mask is just a boolean numpy array.

    They are typically used to represent 'selections' for various operations such as
    when coloring a region of an image.
    """

    def __new__(cls, arr):
        return np.asarray(arr).view(cls)

    def __mul__(self, other):

        try:
            return np.logical_and(self, other)
        except ValueError:

            mask = self.copy()
            size = np.prod(other.shape)

            if mask[mask].shape == size:
                mask[mask] = other.reshape(size)
                return mask

            raise

    def __neg__(self):
        return np.logical_not(self)

    def __sub__(self, other):
        return np.logical_and(self, np.logical_not(other))

    def __rsub__(self, other):
        return np.logical_and(other, np.logical_not(self))

    @classmethod
    def empty(cls, *shape):
        """Return an empty mask with the given shape.

        Example
        -------
        >>> from arlunio.mask import Mask
        >>> Mask.empty(3, 4)
        Mask([[False, False, False, False],
              [False, False, False, False],
              [False, False, False, False]])
        """

        if len(shape) == 1 and isinstance(shape[0], tuple):
            return cls(np.full(shape[0], False))

        return cls(np.full(shape, False))

    @classmethod
    def full(cls, *shape):
        """Return a full mask with the given shape.

        Example
        -------
        >>> from arlunio.mask import Mask
        >>> Mask.full(3, 4)
        Mask([[ True,  True,  True,  True],
              [ True,  True,  True,  True],
              [ True,  True,  True,  True]])
        """

        if len(shape) == 1 and isinstance(shape[0], tuple):
            return cls(np.full(shape[0], True))

        return cls(np.full(shape, True))


@ar.definition
def Empty(width: int, height: int) -> Mask:
    """An empty mask.

    Example
    -------
    >>> from arlunio.mask import Empty
    >>> empty = Empty()
    >>> empty(width=4, height=3)
    Mask([[False, False, False, False],
          [False, False, False, False],
          [False, False, False, False]])
    """
    return Mask.empty(height, width)


@ar.definition
def Full(width: int, height: int) -> Mask:
    """An full mask.

    Example
    -------
    >>> from arlunio.mask import Full
    >>> full = Full()
    >>> full(width=4, height=3)
    Mask([[ True,  True,  True,  True],
          [ True,  True,  True,  True],
          [ True,  True,  True,  True]])
    """
    return Mask.full(height, width)


@ar.definition(operation=ar.Defn.OP_ADD)
def MaskAdd(
    width: int, height: int, *, a: ar.Defn[Mask] = None, b: ar.Defn[Mask] = None
) -> Mask:
    """Add any two mask producing definitions together.

    The resulting defintion will return a mask that is :code:`True` if a given point
    is :code:`True` in either :code:`a` or :code:`b`.

    Attributes
    ----------
    a:
        The first mask
    b:
        The second mask
    """

    return a(width=width, height=height) + b(width=width, height=height)


@ar.definition(operation=ar.Defn.OP_SUB)
def MaskSub(
    width: int, height: int, *, a: ar.Defn[Mask] = None, b: ar.Defn[Mask] = None
) -> Mask:
    """Subtract one mask away from another mask.

    The resulting definition will return a mask that is :code:`True` only if a given
    point is in :code:`a` **and not** in :code:`b`.

    .. note::

       You will get very different results depending on which way around you set
       :code:`a` and :code:`b`!

    Attributes
    ----------
    a:
        The first "base" mask
    b:
        The second mask that defines the region to remove from :code:`a`
    """
    return a(width=width, height=height) - b(width=width, height=height)


@ar.definition(operation=ar.Defn.OP_MUL)
def MaskMul(
    width: int, height: int, *, a: ar.Defn[Mask] = None, b: ar.Defn[Mask] = None
) -> Mask:
    """Muliply any two mask producing definitions together.

    The resulting definition will return a mask that is :code:`True` only if a given
    point is in both :code:`a` and :code:`b`.

    Attributes
    ----------
    a:
        The first mask
    b:
        The second mask
    """
    return a(width=width, height=height) * b(width=width, height=height)


def any_(*args: Union[bool, np.ndarray, Mask]) -> Mask:
    """Given a number of conditions, return :code:`True` if any of the conditions
    are true.

    This function is implemented as a thin wrapper around numpy's
    :data:`numpy:numpy.logical_or` function so that it can take an arbitrary number of
    inputs. This also means that this function will accept arrays of differing sizes,
    assuming that they can be broadcasted to a common shape.

    Parameters
    ----------
    args:
      A number of boolean conditions, a condition can either be a single boolean value
      or a numpy array of boolean values.

    Examples
    --------

    >>> import arlunio.mask as mask
    >>> mask.any_(True, False, False)
    Mask(True)
    >>> mask.any_(False, False, False, False)
    Mask(False)

    If the arguments are boolean numpy arrays, then the any condition is applied
    element-wise

    >>> import numpy as np
    >>> x1 = np.array([True, False, True])
    >>> x2 = np.array([False, False, True])
    >>> x3 = np.array([False, True, False])
    >>> mask.any_(x1, x2, x3)
    Mask([ True,  True,  True])

    The arguments can be any mixture of booleans, arrays and masks.

    >>> mask.any_(False, Mask([True, False]), np.array([[False, True], [True, False]]))
    Mask([[ True,  True],
          [ True, False]])


    See Also
    --------

    :doc:`numpy:user/basics.broadcasting`
       Numpy documentation on broadcasting.

    :doc:`numpy:user/theory.broadcasting`
       Further background on broadcasting.

    :data:`numpy:numpy.logical_or`
       Reference documentation on the :code:`numpy.logical_or` function
    """
    return Mask(functools.reduce(np.logical_or, args))


def all_(*args: Union[bool, np.ndarray, Mask]) -> Mask:
    """Given a number of conditions, return :code:`True` only if **all**
    of the given conditions are true.

    This function is implemented as a thin wrapper around numpy's
    :data:`numpy:numpy.logical_and` function so that it can take an arbitrary number of
    inputs. This also means that this function will accept arrays of differing sizes,
    assuming that they can be broadcasted to a common shape.

    Parameters
    ----------
    args:
      A number of boolean conditions, a conditon can either be a single boolean value,
      or a numpy array of boolean values.

    Examples
    --------

    >>> import arlunio.mask as mask
    >>> mask.all_(True, True, True)
    Mask(True)
    >>> mask.all_(True, False, True, True)
    Mask(False)

    If the arguments are boolean numpy arrays, then the any condition is applied
    element-wise

    >>> import numpy as np
    >>> x1 = np.array([True, False, True])
    >>> x2 = np.array([False, False, True])
    >>> x3 = np.array([False, True, True])
    >>> all_(x1, x2, x3)
    Mask([False, False,  True])

    Arugments can be any mixture of booleans, masks and numpy arrays.

    >>> mask.all_(True, Mask([True, False]), np.array([[False, True], [True, False]]))
    Mask([[False, False],
          [ True, False]])


    See Also
    --------

    :doc:`numpy:user/basics.broadcasting`
       Numpy documentation on broadcasting.

    :doc:`numpy:user/theory.broadcasting`
       Further background on broadcasting.

    :data:`numpy:numpy.logical_and`
       Reference documentation on the :code:`logical_and` function.
    """
    return Mask(functools.reduce(np.logical_and, args))



import inspect
import textwrap
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from sphinx.application import Sphinx

import arlunio
from .image import register as register_image
from .notebook import register as register_notebook

# fmt: off
TEMPLATE = [
    ".. list-table::",
    "   :widths: 5 20",
    "   :header-rows: 0",
    "",
]
# fmt: on


def _document_inputs(defn: arlunio.Defn, lines: List[str]):
    """Given a definition, document any inputs."""

    inputs = []

    for name, value in defn.inputs().items():
        dtype = value.dtype

        if hasattr(value.dtype, "__name__"):
            dtype = dtype.__name__

        inputs.append(f":code:`{name}: {dtype}`")

    if len(inputs) > 0:
        lines.append(textwrap.indent("* - **Inputs:**", " " * 3))
        lines.append(textwrap.indent(" ".join(inputs), " " * 5 + "- "))


def _document_bases(defn: arlunio.Defn, lines: List[str]):
    """Given a definition, document any definitions it is derived from."""

    defns = []

    for name, value in defn.bases().items():

        name = value.__name__
        mod = value.__module__

        defns.append(f":class:`{name} <{mod}.{name}>`")

    if len(defns) > 0:
        lines.append(textwrap.indent("* - **Bases:**", " " * 3))
        lines.append(textwrap.indent(" ".join(defns), " " * 5 + "- "))


def _document_produces(defn: arlunio.Defn, lines: List[str]):
    """Given a definition and it declares with it produces, document it."""

    if defn.produces() == Any:
        return

    result = defn.produces()

    name = result.__name__
    mod = result.__module__

    lines.append(textwrap.indent("* - **Produces:**", " " * 3))
    lines.append(textwrap.indent(f":class:`{name} <{mod}.{name}>`", " " * 5 + "- "))


def _document_inheritance(defn: arlunio.Defn) -> Optional[List[str]]:
    """Given a definition, link back to any definitions it derives from."""

    lines = []

    _document_inputs(defn, lines)
    _document_bases(defn, lines)
    _document_produces(defn, lines)

    if len(lines) == 0:
        return None

    return list(TEMPLATE) + lines + ["", ""]


def _process_docstring(
    app: Sphinx, what: str, name: str, obj: Any, options: Any, lines: List[str]
) -> None:

    if inspect.isclass(obj) and issubclass(obj, arlunio.Defn):
        inherits = _document_inheritance(obj)

        if inherits is not None:

            for line in reversed(inherits):
                lines.insert(0, line)


def setup(app: Sphinx) -> Dict[str, Any]:

    app.setup_extension("sphinx.ext.autodoc")
    app.connect("autodoc-process-docstring", _process_docstring)

    register_image(app)
    register_notebook(app)

    return {"version": arlunio.__version__, "parallel_read_safe": True}



import argparse
import collections
import logging
import sys
import textwrap
import traceback

import pkg_resources

from ._interface import build_command_parser
from arlunio._version import __version__

logger = logging.getLogger(__name__)

_LogConfig = collections.namedtuple("LogConfig", "level,fmt")
_LOG_LEVELS = [
    _LogConfig(level=logging.INFO, fmt="%(message)s"),
    _LogConfig(level=logging.DEBUG, fmt="[%(levelname)s]: %(message)s"),
    _LogConfig(level=logging.DEBUG, fmt="[%(levelname)s][%(name)s]: %(message)s"),
]


def _setup_logging(verbose: int, quiet: bool) -> None:
    """Setup the logging system according to the given args."""

    if quiet:
        return

    verbose = 0 if verbose < 0 else verbose

    try:
        conf = _LOG_LEVELS[verbose]
        others = False
    except IndexError:
        conf = _LOG_LEVELS[-1]
        others = True

    root = logging.getLogger()
    root.setLevel(conf.level)

    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter(conf.fmt))

    if not others:
        console.addFilter(logging.Filter("arlunio"))

    root.addHandler(console)


def build_dummy_command_parser(
    name: str, parent: argparse._SubParsersAction, err
) -> None:
    """Build a dummy command to report an error encountered while loading the real
    one."""

    e_type, e_val, tb = err

    message = "Command unavailable: " + str(e_val)
    parser = parent.add_parser(name, help=message)
    parser.add_argument("collector", nargs=argparse.REMAINDER)

    header = f"\nCommand Unavailable: {name}\n\n"

    body = """\
    This is either because your environment does not have the dependencies
    available required by this command or there is an issue with the command
    itself.\n
    """
    body = textwrap.dedent(body)

    body += f"\t{e_type.__name__}: {e_val}\n"

    if issubclass(e_type, ImportError):
        extra = """\

        Since this is an import error it's probably an issue with your environment
        please make sure you have all the required libraries installed.
        """
        body += textwrap.dedent(extra)

    message = header + body

    def run(args: argparse.Namespace):
        logger.info(message)
        logger.debug("\n%s", "\n".join(traceback.format_exception(e_type, e_val, tb)))

    parser.set_defaults(run=run)


def _register_commands(parent: argparse._SubParsersAction, entry_point: str):
    """Given the name of an entrypoint expose all the commands under the given parent.

    In the case where a command cannot be loaded, this function will put a dummy command
    in its place so that we can gracefully handle it and inform the user.
    """
    commands = sorted(
        list(pkg_resources.iter_entry_points(entry_point)), key=lambda c: c.name
    )
    for cmd in commands:
        try:
            command = cmd.load()
            build_command_parser(cmd.name, command, parent)
        except Exception:
            info = sys.exc_info()
            build_dummy_command_parser(cmd.name, parent, info)


# Top level command line parser
_cli = argparse.ArgumentParser(description="Arlunio's command line interface")
_cli.add_argument(
    "-q", "--quiet", help="disable all console output", action="store_true"
)
_cli.add_argument("--version", help="show version info and exit", action="store_true")
_cli.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=0,
    help="increase output verbosity, repeatable e.g. -v, -vv, -vvv, ...",
)

_commands = _cli.add_subparsers(title="commands")
_register_commands(_commands, "arlunio.cli.commands")


def main():
    args = _cli.parse_args()

    if args.version:
        logger.info("arlunio v%s", __version__)
        return 0

    if hasattr(args, "run"):
        _setup_logging(args.verbose, args.quiet)
        return args.run(args)

    _cli.print_help()
    return 0



from .camera import SimpleCamera
from .camera import SimpleSampler
from .camera import UniformSampler
from .data import Rays
from .data import ScatterPoint
from .material import Gradient
from .material import LambertianDiffuse
from .material import NormalMap
from .object import Sphere
from .render import ClayRenderer
from .render import MaterialRenderer
from .render import SampledRenderer
from .render import ZDepthRenderer

__all__ = [
    "ClayRenderer",
    "Gradient",
    "LambertianDiffuse",
    "MaterialRenderer",
    "NormalMap",
    "Rays",
    "ScatterPoint",
    "SimpleCamera",
    "SampledRenderer",
    "SimpleSampler",
    "Sphere",
    "UniformSampler",
    "ZDepthRenderer",
]



import logging

import numpy as np

import arlunio as ar
from .data import Rays
from .data import ScatterPoint
from arlunio.mask import all_
from arlunio.math import dot
from arlunio.math import normalise


@ar.definition
def Sphere(
    rays: Rays, t_min: float, t_max: float, *, center=None, radius=0.5
) -> ScatterPoint:
    """A sphere."""

    logger = logging.getLogger(__name__)

    center = np.array([0, 0, -1]) if center is None else center

    oc = rays.origin - center

    a = dot(rays.direction, rays.direction)
    b = dot(oc, rays.direction)
    c = dot(oc, oc) - radius * radius

    logger.debug("a: %s, b: %s, c: %s", a.shape, b.shape, c.shape)

    # Determine which rays intersect
    disc = b * b - a * c
    hits = disc > 0

    logger.debug(
        "disc: %s, hits: %s, #hits: %s",
        disc.shape,
        hits.shape,
        hits[hits == True].shape,
    )

    a, b, root = a[hits], b[hits], np.sqrt(disc[hits])

    # Determine the smallest value of t for those intersections
    t1 = (-b - root) / a
    t2 = (-b + root) / a

    # Assume t1 is the correct value to choose
    t = t1

    # But where t1 < t_min, replace it with t2 if applicable
    mask = all_(t1 < t_min, t2 > t_min)
    t[mask] = t2[mask]

    # Now apply the bounds to t to get the actucal hits
    mask = all_(t > t_min, t < t_max[hits])
    hits[hits] = mask
    t = t[mask]

    # Determine the points of intersection and surface normals at t
    p = rays[hits].at(t)
    normals = normalise(p - center)

    # Finally determine which rays are hitting the outside surface of the sphere
    front_face = dot(rays.direction[hits], normals) < 0
    normals[front_face != True] = -normals[front_face != True]

    return ScatterPoint(hits, t, p, normals, front_face)



"""Helpers and utilities for writing tests."""
import hypothesis.strategies as st
import numpy as np
from hypothesis.extra import numpy

pve_num = st.floats(min_value=1, max_value=1e6)
real_num = st.floats(min_value=-1e6, max_value=1e6)

dimension = st.integers(min_value=4, max_value=512)

mask = numpy.arrays(
    dtype=np.bool_, shape=st.tuples(dimension, dimension), fill=st.booleans()
)



import argparse
import inspect
import logging
from typing import Any
from typing import Dict
from typing import List

import attr

logger = logging.getLogger(__name__)


def _build_bool_option(parameter: inspect.Parameter) -> Dict[str, Any]:
    """Provide the argparse options for a boolean option."""
    params = {"action": "store_false" if parameter.default is True else "store_true"}

    return params


def _build_float_option(parameter: inspect.Parameter) -> Dict[str, Any]:
    """Provide the argparse options for a float option."""
    params = {
        "type": float,
        "default": None if parameter.default == parameter.empty else parameter.default,
    }

    return params


def _build_int_option(parameter: inspect.Parameter) -> Dict[str, Any]:
    """Provides the argparse options for an integer option."""
    params = {
        "type": int,
        "default": None if parameter.default == parameter.empty else parameter.default,
    }

    return params


_STD_TYPES = {
    bool: _build_bool_option,
    float: _build_float_option,
    int: _build_int_option,
}


@attr.s(auto_attribs=True)
class CliOption:
    """Represents a command line option."""

    name: str
    """The name of the option."""

    description: str = None
    """Description about what the option controls."""

    args: Dict[str, Any] = attr.Factory(dict)
    """Extra arguments required to express this option in argparse."""

    @classmethod
    def fromparam(cls, parameter: inspect.Parameter):
        """Construct an instance of a cli option from a type annotation."""
        params = {"name": "--" + parameter.name.replace("_", "-")}
        builder = _STD_TYPES.get(parameter.annotation, None)

        if builder is None:
            raise TypeError(f"Unsupported cli option type: {parameter.annotation}")

        params["args"] = builder(parameter)
        return cls(**params)

    @classmethod
    def frommethod(cls, f):
        """Return a list of command line options, constructed from the given
        function."""
        options = []

        for param in inspect.signature(f).parameters.values():
            # Ignore the 'self' parameter in classes.
            if param.name == "self":
                continue

            options.append(cls.fromparam(param))

        return options


@attr.s(auto_attribs=True)
class CliCommand:
    """Class that represents a command on the command line."""

    name: str
    """The name of the command, this will be used to invoke it on the command line."""

    summary: str = None
    """Short, one line description of the command."""

    description: str = None
    """Extended, detailed description of the command and its options."""

    options: List[CliOption] = attr.Factory(list)
    """List of command line options, these will be exposed as flags on the command."""

    @classmethod
    def fromcmd(cls, name: str, cmd):
        """Create an instance of a command, based on a single command class."""

        if not hasattr(cmd, "run") or not callable(cmd.run):
            raise TypeError("Missing expected method: 'run'")

        docstring = inspect.getdoc(cmd)
        summary = "..." if docstring is None else docstring.split("\n")[0]

        params = {
            "name": name,
            "options": CliOption.frommethod(cmd.run),
            "summary": summary,
        }

        return cls(**params)


def build_command_parser(
    name: str, command, parent: argparse._SubParsersAction
) -> None:
    """Given a command defintion, define its command line interface.

    :param name: The name of the command, as it should be invoked from the command line.
    :param command: The class that defines the command's implementation.
    :param parent: The parent cli entity the command should hang off of.
    """
    cmd = CliCommand.fromcmd(name, command)
    parser = parent.add_parser(
        cmd.name, help=cmd.summary, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    names = []
    for option in cmd.options:
        names.append(option.name[2:])
        parser.add_argument(option.name, **option.args)

    def run(args: argparse.Namespace):
        params = {opt: getattr(args, opt) for opt in names}
        logger.debug("Parsed arguments: %s", params)

        instance = command()
        instance.run(**params)

    parser.set_defaults(run=run)



import logging

import numpy as np
import PIL.Image as Image

import arlunio as ar
from .camera import SimpleCamera
from .data import Rays
from .data import ScatterPoint
from .material import Gradient
from .material import LambertianDiffuse
from arlunio.math import clamp


@ar.definition
def ZDepthRenderer(width: int, height: int, *, camera=None, objects=None):
    """A renderer that returns a z-depth pass."""

    camera = SimpleCamera() if camera is None else camera
    objects = [] if objects is None else objects

    rays = camera(width=width, height=height)
    scatter = ScatterPoint.new(rays)

    for obj in objects:
        test = obj(rays=rays, t_min=0, t_max=scatter.t)
        scatter.merge(test)

    depth = np.abs(scatter.p[:, 2] - camera.origin[2])
    depth[scatter.t == ScatterPoint.MAX_FLOAT] = np.max(depth)

    vs = np.array((depth / np.max(depth)) * 255, dtype=np.uint8)
    vs = vs.reshape(height, width)

    return Image.fromarray(vs, "L")


@ar.definition
def ClayRenderer(
    width: int,
    height: int,
    *,
    background=None,
    color="lightgrey",
    camera=None,
    objects=None,
    bounces=8,
):
    """A simple clay renderer."""

    background = Gradient() if background is None else background
    camera = SimpleCamera() if camera is None else camera
    objects = [] if objects is None else objects

    rays = camera(width=width, height=height)
    scatter = ScatterPoint.new(rays)
    mat = LambertianDiffuse(color=color)
    color = background(scatter=scatter)

    depth = 1
    mask = np.full((scatter.hit.shape), True)

    while depth < bounces:

        for obj in objects:
            test = obj(rays=rays, t_min=0.001, t_max=scatter.t)
            scatter.merge(test)

        # If there were no hits, then there's nothing to do
        if not np.any(scatter.hit):
            break

        # Update the mask to reflect where we have intersections.
        mask[mask] = scatter.hit

        matcol, rays = mat(scatter=scatter[scatter.hit])
        color[mask] *= matcol

        # Increasee the bounce counter and run again
        depth += 1
        scatter = ScatterPoint.new(rays)

    return color


@ar.definition
def MaterialRenderer(
    width: int, height: int, *, objects=None, bounces=8, camera=None, background=None
):
    """A renderer capable of rendering materials."""

    logger = logging.getLogger(__name__)

    background = Gradient() if background is None else background
    camera = SimpleCamera() if camera is None else camera
    objects = [] if objects is None else objects

    rays = camera(width=width, height=height)
    scatter = ScatterPoint.new(rays)
    color = background(scatter=scatter)

    depth = 1
    mask = np.full((scatter.hit.shape), True)

    materials = [mat for _, mat in objects]

    logger.info("Starting render.")

    while depth < bounces:

        matmap = np.full((scatter.hit.shape), -1)

        logger.debug(" Depth: %i ".center(80, "-"), depth)
        logger.debug("Mask: %s", mask.shape)
        logger.debug("Material Map: %s", matmap.shape)
        logger.debug("%s", scatter)

        for idx, (obj, _) in enumerate(objects):
            logger.debug(" Obj: %i ".center(80, "-"), idx)
            test = obj(rays=rays, t_min=0.001, t_max=scatter.t)

            # Record which material should be used for these intersections.
            matmap[test.hit] = idx
            scatter.merge(test)

        if not np.any(scatter.hit):
            logger.debug("Nothing hit, stopping")
            break

        matids = np.unique(matmap[matmap != -1])
        logger.debug("-" * 80)
        logger.debug("Mat Ids: %s", matids)
        logger.debug("#hits: %s", scatter.hit[scatter.hit == True].shape)

        mask[mask] = scatter.hit
        origin = scatter.p[scatter.hit]
        directions = scatter.normal[scatter.hit]

        for idx in matids:
            logger.debug(" Mat: %i ".center(80, "-"), idx)
            mat = materials[idx]

            # Only select the intersections that correspond with this material
            mat_mask = matmap[scatter.hit] == idx

            scatter_mask = np.array(scatter.hit)
            scatter_mask[scatter.hit] = mat_mask

            color_mask = np.array(mask)
            color_mask[color_mask] = scatter_mask[scatter.hit]

            logger.debug("Mat Mask: %s", mat_mask.shape)
            logger.debug("Scatter Mask: %s", scatter_mask.shape)
            logger.debug("Color Mask: %s", color_mask.shape)

            mat_scatter = scatter[scatter_mask]
            logger.debug("Mat Scatter: %s", mat_scatter)

            matcol, matrays = mat(scatter=mat_scatter)
            color[color_mask] *= matcol

            origin[mat_mask] = matrays.origin
            directions[mat_mask] = matrays.direction

        depth += 1

        rays = Rays(origin, directions)
        scatter = ScatterPoint.new(rays)

    return color


@ar.definition
def SampledRenderer(width: int, height: int, *, kernel=None, samples=10):
    """A renderer is responsible for orchestrating the entire process."""

    if kernel is None:
        raise ValueError("Missing renderer kernel")

    s = samples
    n = width * height
    scale = 1.0 / s

    cols = np.zeros((n, 3))
    for _ in range(s):
        cols += kernel(width=width, height=height) * scale

    # Gamma correction...
    cols = np.sqrt(cols)

    r = clamp(cols[:, 0], 0, 0.99999) * 256
    g = clamp(cols[:, 1], 0, 0.99999) * 256
    b = clamp(cols[:, 2], 0, 0.99999) * 256

    color = np.array(np.dstack([r, g, b])[0], dtype=np.uint8)
    color = color.reshape(height, width, 3)

    return Image.fromarray(color, "RGB")



import numpy as np
import numpy.random as npr

import arlunio as ar
from .data import Rays


@ar.definition
def SimpleSampler(width: int, height: int):
    """The simplest sampler."""

    u = np.full((height, width), np.linspace(0, 1, width))

    v = np.linspace(1, 0, height).reshape(1, height).transpose()
    v = np.full((height, width), v)

    return np.dstack([u, v]).reshape(width * height, 2)


@ar.definition
def UniformSampler(width: int, height: int):
    """Generate samples according to the uniform distribution"""

    n = width * height
    u = np.full((height, width), np.linspace(0, 1, width))

    v = np.linspace(1, 0, height).reshape(1, height).transpose()
    v = np.full((height, width), v)

    uv = np.dstack([u, v]).reshape(n, 2)
    uv[:, 0] += npr.rand(n) / width
    uv[:, 1] += npr.rand(n) / height

    return uv


@ar.definition
def SimpleCamera(
    width: int,
    height: int,
    *,
    origin=np.array([0.0, 0.0, 0.0]),
    scale=2.0,
    focal_length=1.0,
    sampler=None
) -> Rays:
    """A simple camera that generates rays to be cast into the scene.

    Attributes
    ----------
    focal_length:
        Something something focal length
    origin:
        Set the position of the camera in space. If not set this will default to
        :math:`(0, 0, 0)`
    sampler:
        The sampler the camera should use in order to generate rays. If not set this
        will default to an instance of :class:`arlunio.raytrace.UniformSampler`
    scale:
        Something something scale.
    """

    # Pick sensible defaults if some attributes are not set
    origin = np.array(origin)
    sampler = UniformSampler() if sampler is None else sampler

    ratio = width / height

    # Not entirely sure how to describe the effect these have on the final image...
    horizontal = np.array([scale * ratio, 0.0, 0.0])
    vertical = np.array([0.0, scale, 0.0])

    uv = sampler(width=width, height=height)
    lower_left = (
        origin - (horizontal / 2) - (vertical / 2) - np.array([0, 0, focal_length])
    )

    n = uv.shape[0]

    hs = np.einsum("n,np->np", uv[:, 0], horizontal.reshape(1, 3))
    vs = np.einsum("n,np->np", uv[:, 1], vertical.reshape(1, 3))
    ll = np.full((n, 3), lower_left)

    origin = np.full((n, 3), origin)
    directions = ll + hs + vs

    return Rays(origin, directions)



import os

from setuptools import find_packages
from setuptools import setup

info = {}
version = os.path.join("arlunio", "_version.py")

with open(version) as f:
    exec(f.read(), info)


def readme():
    with open("README.md") as f:
        return f.read()


required = ["attrs", "appdirs", "ipython", "numpy", "Pillow>=6.1.0"]
extras = {
    "dev": [
        "black",
        "flake8",
        "hypothesis",
        "jupyterlab",
        "pre-commit",
        "pytest",
        "pytest-cov",
        "sphinx-autobuild",
        "sphinx_rtd_theme",
        "sphobjinv",
        "towncrier",
        "tox",
    ],
    "doc": ["sphinx", "nbformat"],
    "examples": ["jupyterlab"],
    "testing": ["hypothesis"],
}

extras["all"] = list(
    {item for name, items in extras.items() if name != "dev" for item in items}
)

setup(
    name="arlunio",
    version=info["__version__"],
    project_urls={
        "Documentation": "https://www.arlun.io/docs/",
        "Source": "https://github.com/swyddfa/arlunio",
        "Tracker": "https://github.com/swyddfa/arlunio/issues",
    },
    description="Drawing and animating with a blend of Python and mathematics.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="Swyddfa Developers",
    author_email="swyddfa.dev@gmail.com",
    license="MIT",
    packages=find_packages(".", exclude=["tests*"]),
    package_data={
        "arlunio.cli.scripts": ["*.sh"],
        "arlunio.tutorial": ["*.ipynb", "**/*.ipynb"],
        "arlunio.tutorial.solutions": ["*.py"],
    },
    python_requires=">=3.6",
    install_requires=required,
    extras_require=extras,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Multimedia :: Graphics",
    ],
    entry_points={
        "console_scripts": ["arlunio = arlunio.cli.__main__:main"],
        "sphinx.builders": ["nbtutorial = arlunio.doc", "nbgallery = arlunio.doc"],
        "arlunio.cli.commands": [
            "repl = arlunio.cli.repl:Repl",
            "tutorial = arlunio.cli.tutorial:Tutorial",
        ],
    },
)



import numpy as np

import arlunio as ar
import arlunio.mask as mask
import arlunio.math as math


@ar.definition
def Circle(x: math.X, y: math.Y, *, xc=0, yc=0, r=0.8, pt=None) -> mask.Mask:
    """
    .. arlunio-image:: Basic Circle
       :align: right

       ::

          from arlunio.shape import Circle
          from arlunio.image import fill

          circle = Circle()
          image = fill(circle(width=256, height=256))

    We define a circle using the following equality.

    .. math::

       (x - x_c)^2 + (y - y_c)^2 = r^2

    Attributes
    ----------
    xc:
        Corresponds with the :math:`x_c` variable in the equation above and defines the
        :math:`x`-coordinate of the circle's center.
    yc:
        Corresponds with the :math:`y_c` variable in the equation above and defines the
        :math:`y`-coordinate of the circle's center.
    r:
        Corresponds with the :math:`r` variable in the equation above and defines the
        radius of the circle.
    pt:
        If :code:`None`, then all points within the radius of the circle will be
        considered to be part of it. If this is set to some positive number then all
        points between radii :code:`(1 - pt) * r` and :code:`(1 + pt) * r` will be
        considered part of the circle.

    Examples
    --------

    .. arlunio-image:: Target
       :include-code:
       :gallery: examples

       Combining a few circles we're able to draw a target::

          import arlunio as ar

          from arlunio.shape import Circle
          from arlunio.image import fill

          @ar.definition
          def Target(width: int, height: int):
              image = None
              parts = [
                  (Circle(pt=0.02), "#000"),
                  (Circle(r=0.75, pt=0.12), "#f00"),
                  (Circle(r=0.6, pt=0.05), "#f00"),
                  (Circle(r=0.4), "#f00"),
              ]

              for part, color in parts:
                  image = fill(
                      part(width=width, height=height), foreground=color, image=image
                  )

              return image

          target = Target()
          image = target(width=1920, height=1080)

    Making use of the :code:`xc` and :code:`yc` attributes we can produce an
    approximation of the olympics logo

    .. arlunio-image:: Olympic Rings
       :include-code:
       :gallery: examples

       ::

          import arlunio as ar

          from arlunio.shape import Circle
          from arlunio.image import fill

          @ar.definition
          def OlympicRings(width: int, height: int, *, spacing=0.5, pt=0.025):

              dy = spacing / 4
              dx = spacing / 2
              args = {"scale": 0.5, "r": spacing, "pt": pt}

              image = None
              rings = [
                  (Circle(yc=dy, xc=-(2.2 * dx), **args), "#0ff"),
                  (Circle(yc=dy, **args), "#000"),
                  (Circle(yc=dy, xc=(2.2 * dx), **args), "#f00"),
                  (Circle(yc=-dy, xc=-(1.1 * dx), **args), "#ff0"),
                  (Circle(yc=-dy, xc=(1.1 * dx), **args), "#0f0")
              ]

              for ring, color in rings:
                  image = fill(
                      ring(width=width, height=height), foreground=color, image=image
                  )

              return image

          rings = OlympicRings()
          image = rings(width=1920, height=1080)
    """

    x = (x - xc) ** 2
    y = (y - yc) ** 2
    circle = np.sqrt(x + y)

    if pt is None:
        return mask.Mask(circle < r ** 2)

    inner = (1 - pt) * r ** 2
    outer = (1 + pt) * r ** 2

    return mask.all_(inner < circle, circle < outer)


@ar.definition
def Ellipse(x: math.X, y: math.Y, *, xc=0, yc=0, a=2, b=1, r=0.8, pt=None) -> mask.Mask:
    """
    .. arlunio-image:: Simple Ellipse
       :align: right

       ::

          from arlunio.shape import Ellipse
          from arlunio.image import fill

          ellipse = Ellipse(r=0.6)
          image = fill(ellipse(width=256, height=256))

    An ellipse can be defined using the following equality.

    .. math::

       \\left(\\frac{x - x_c}{a}\\right)^2 +
       \\left(\\frac{y - y_c}{b}\\right)^2 = r^2

    Attributes
    ----------
    xc:
        Corresponds with the :math:`x_c` variable in the equation above and defines the
        :math:`x`-coordinate of the ellipse's center.
    yc:
        Corresponds with the :math:`y_c` variable in the equation above and defines the
        :math:`y`-coordinate of the ellipse's center.
    r:
        Corresponds with the :math:`r` variable in the equation above and controls the
        overall size of the ellipse.
    a:
        Corresponds with the :math:`a` variable in the equation above and controls the
        width of the ellipse.
    b:
        Corresponds with the :math:`b` variable in the equation above and controls the
        height of the ellipse.
    pt:
        If :code:`None` then all points within the radius of the ellipse will be
        considered to be part of it. If this is set to some positive number then all
        points between radii :code:`(1 - pt) * r` and :code:`(1 + pt) * r` will be
        considered part of the ellipse.

    Examples
    --------

    :code:`a` and :code:`b` together determine the overall shape of the ellipse.
    Increasing the value of :code:`a` will stretch the ellipse width wise, increasing
    :code:`b` has a similar effect for the height. It's worth noting that it's the
    ratio of these 2 values rather than their absolute values that has a greater effect
    on the shape of the ellipse. If :code:`a = b` then the equation simplifies to that
    of a circle

    .. arlunio-image:: Ellipse Demo
       :include-code:

       ::

          import arlunio as ar

          from arlunio.shape import Ellipse
          from arlunio.image import fill

          @ar.definition
          def EllipseDemo(width: int, height: int):
              image = None
              ellipses = [
                  Ellipse(xc=-0.5, yc=-0.5, a=0.5, b=0.5, r=0.4),
                  Ellipse(yc=-0.5, a=1, b=0.5, r=0.4),
                  Ellipse(xc=0.5, yc=-0.5, a=2, b=0.5, r=0.4),
                  Ellipse(a=1, b=1, r=0.4),
                  Ellipse(xc=0.5, yc=0.5, a=2, b=2, r=0.4),
                  Ellipse(xc=-0.5, a=0.5, b=1, r=0.4),
                  Ellipse(xc=-0.5, yc=0.5, a=0.5, b=2, r=0.4)
              ]

              for ellipse in ellipses:
                  image = fill(ellipse(width=1920, height=1080), image=image)
              return image

          demo = EllipseDemo()
          image = demo(width=1920, height=1080)

    .. arlunio-image:: Atom
       :include-code:
       :gallery: examples

       Playing around with the values and the coordinate inputs it's possible to
       draw something that looks like a diagram of an atom::

          import arlunio as ar

          from arlunio.shape import Ellipse
          from arlunio.math import X, Y
          from arlunio.image import fill

          @ar.definition
          def Atom(x: X, y: Y):
              image = None

              ellipses = [
                  (Ellipse(a=1.5, b=0.5, pt=0.005), x, y),
                  (Ellipse(a=1.5, b=0.5, r=1, pt=0.005), x + y, y - x),
                  (Ellipse(a=0.5, b=1.5, pt=0.005), x, y),
                  (Ellipse(a=1.5, b=0.5, r=1, pt=0.005), x - y, x + y),
                  (Ellipse(a=1, b=1, r=0.15), x, y)
              ]

              for ellipse, ex, ey in ellipses:
                  image = fill(ellipse(x=ex, y=ey), image=image)

              return image

          atom = Atom()
          image = atom(width=1920, height=1080)
    """

    x = (x - xc) ** 2
    y = (y - yc) ** 2

    a = a ** 2
    b = b ** 2

    ellipse = np.sqrt(x / a + y / b)

    if pt is None:
        return mask.Mask(ellipse < r * r)

    inner = (1 - pt) * r ** 2
    outer = (1 + pt) * r ** 2

    return mask.all_(inner < ellipse, ellipse < outer)


@ar.definition
def SuperEllipse(
    x: math.X, y: math.Y, *, xc=0, yc=0, a=1, b=1, n=3, r=0.8, m=None, pt=None
) -> mask.Mask:
    """
    .. arlunio-image:: SuperEllipse
       :align: right

       ::

          from arlunio.shape import SuperEllipse
          from arlunio.image import fill

          ellipse = SuperEllipse()
          image = fill(ellipse(width=256, height=256))

    We define a `SuperEllipse`_ by the following equality.

    .. math::

       \\left|\\frac{(x - x_c)}{a}\\right|^n + \\left|\\frac{(y - y_c)}{b}\\right|^m = r

    Attributes
    ----------
    xc:
        Corresponds with the :math:`x_c` variable in the equation above and defines the
        :math:`x`-coordinate of the center of the super ellipse.
    yc:
        Corresponds with the :math:`y_c` variable in the equation above and defines the
        :math:`y` -coordinate of the center of the super ellipse.
    r:
        Corresponds with the :math:`r` variable in the equation above and controls the
        size of the super ellipse.
    a:
        Corresponds with the :math:`a` variable in the equation above and controls the
        width of the super ellipse.
    b:
        Corresponds with the :math:`b` variable in the equation above and controls the
        height of the super ellipse.
    n:
        Corresponds with the :math:`n` variable in the equation above and controls the
        profile of the curve far from :math:`x = 0`
    m:
        Corresponds with the :math:`m` variable in the equation above and controls the
        profile of the curve close to :math:`x = 0`. If :code:`m = None` (default) then
        it will be set to the value of :code:`n`.
    pt:
        If :code:`None` then all points within the radius of the super ellipse will be
        considered to be part of it. If this is set to some positive number then all
        points between radii :code:`(1 - pt) * r` and :code:`(1 + pt) * r` will be
        considered part of the super ellipse.

    Examples
    --------

    Being a generalisation of the regular :class:`arlunio.shape.Ellipse` definition most
    of the attributes will have a similar effect on the outcome so be sure to check it
    out for additional examples. For the :code:`SuperEllipse` definition the most
    interesting attributes are :code:`n` and :code:`m` greatly affect the shape of the
    super ellipse.

    .. arlunio-image:: SuperEllipse Demo
       :include-code:
       :gallery: examples

       ::

          import arlunio as ar

          from arlunio.shape import SuperEllipse
          from arlunio.image import fill

          @ar.definition
          def SuperEllipseDemo(width: int, height: int):
              image = None
              ellipses = [
                  (SuperEllipse(n=0.5, pt=0.01),'#f00'),
                  (SuperEllipse(n=1, pt=0.01),'#0f0'),
                  (SuperEllipse(n=1.5, pt=0.01), '#00f'),
                  (SuperEllipse(n=2, pt=0.01), '#ff0'),
                  (SuperEllipse(n=3, pt=0.01), '#0ff')
              ]

              for ellipse, color in ellipses:
                  image = fill(
                      ellipse(width=1920, height=1080), foreground=color, image=image
                  )

              return image

          demo = SuperEllipseDemo()
          image = demo(width=1920, height=1080)

    By default if you don't specify a value for :code:`m` it will inherit the value
    assigned to :code:`n`. However if you set :code:`m` to a different value then you
    can get even more interesting results!

    .. arlunio-image:: Eye of Sauron
       :include-code:
       :gallery: examples

       "Eye of Sauron"::

          import arlunio as ar

          from arlunio.shape import SuperEllipse
          from arlunio.image import fill

          @ar.definition
          def Sauron(width: int, height: int):
              image = None
              ellipses = [
                  (SuperEllipse(a=2, n=3, m=0.2, r=0.98),'#f00'),
                  (SuperEllipse(n=2),'#f50'),
                  (SuperEllipse(n=0.1, m=2), '#000'),
              ]

              for ellipse, color in ellipses:
                  image = fill(
                      ellipse(width=1920, height=1080), foreground=color, image=image
                  )

              return image

          eye = Sauron()
          image = eye(width=1920, height=1080)

    .. _SuperEllipse: https://en.wikipedia.org/wiki/Superellipse

    """

    x = x - xc
    y = y - yc

    if m is None:
        m = n

    ellipse = np.abs(x / a) ** n + np.abs(y / b) ** m

    if pt is None:
        return mask.Mask(ellipse < r)

    inner = (1 - pt) * r
    outer = (1 + pt) * r

    return mask.all_(inner < ellipse, ellipse < outer)


@ar.definition
def Square(x: math.X, y: math.Y, *, xc=0, yc=0, size=0.8, pt=None) -> mask.Mask:
    """
    .. arlunio-image:: Simple Square
       :align: right

       A square::

          from arlunio.shape import Square
          from arlunio.image import fill

          square = Square()
          image = fill(square(width=256, height=256))

    Attributes
    ----------
    xc:
        Defines the :math:`x`-coordinate of the square's center
    yc:
        Defines the :math:`y`-coordinate of the square's center
    size:
        Defines the size of the square, sides will have a length of :code:`2 * size`
    pt:
        If :code:`None` then all points within the square's border will be considered to
        be part of it. If set to some positive number then all points within
        :code:`(1 - pt) * size` to :code:`(1 + pt) * size` of the border will be
        considered to be part of the square.

    Example
    -------

    .. arlunio-image:: Square Demo
       :include-code:
       :gallery: examples

       ::

          import arlunio as ar

          from arlunio.shape import Square
          from arlunio.math import X, Y
          from arlunio.image import fill

          @ar.definition
          def SquareDemo(x: X, y: Y):
              image = None
              squares = [
                  (Square(pt=0.01), x, y),
                  (Square(pt=0.01), x + y, x - y),
                  (Square(size=0.39, pt=0.01), x, y),
                  (Square(size=0.39, pt=0.01), x + y, x - y),
                  (Square(size=0.2), x, y),
              ]

              for square, sx, sy in squares:
                  image = fill(square(x=sx, y=sy), image=image)

              return image

          square = SquareDemo()
          image = square(width=1920, height=1080)
    """

    xs = np.abs(x - xc)
    ys = np.abs(y - yc)

    if pt is None:
        return mask.all_(xs < size, ys < size)

    s = (1 - pt) * size
    S = (1 + pt) * size

    inner = mask.all_(xs < s, ys < s)
    outer = mask.all_(xs < S, ys < S)

    return outer - inner


@ar.definition
def Rectangle(
    x: math.X, y: math.Y, *, xc=0, yc=0, size=0.6, ratio=1.618, pt=None
) -> mask.Mask:
    """
    .. arlunio-image:: Simple Rectangle
       :align: right

       A rectangle::

          from arlunio.shape import Rectangle
          from arlunio.image import fill

          rectangle = Rectangle()
          image = fill(rectangle(width=256, height=256))

    Attributes
    ----------
    xc:
        Defines the :math:`x`-coordinate of the center of the rectangle
    yc:
        Defines the :math:`y`-coordinate of the center of the rectangle
    size:
        Defines the area of the rectangle
    ratio:
        Defines the ratio of the width to the height of the rectangle
    pt:
        If :code:`None` then all points within the rectangle's border will be considered
        to be part of it. If set to some positive number then all points within
        :code:`(1 - pt) * size` to :code:`(1 + pt) * size` of the border will be
        considered to be part of the rectangle.

    Examples
    --------

    .. arlunio-image:: Rectangle Demo
       :include-code:

       ::

          import arlunio as ar

          from arlunio.shape import Rectangle
          from arlunio.image import fill

          @ar.definition
          def RectangleDemo(width: int, height: int):
              image = None
              rects = [
                  Rectangle(xc=-1, size=0.4, ratio=0.5),
                  Rectangle(xc=0.25, yc=0.5, size=0.2, ratio=1),
                  Rectangle(xc=0.5, yc=-0.5, size=0.4, ratio=2)
              ]

              for r in rects:
                  image = fill(r(width=width, height=height), image=image)

              return image

          demo = RectangleDemo()
          image = demo(width=1920, height=1080)
    """

    xs = np.abs(x - xc)
    ys = np.abs(y - yc)

    height = np.sqrt(size / ratio)
    width = height * ratio

    if pt is None:
        return mask.all_(xs < width, ys < height)

    w, W = (1 - pt) * width, (1 + pt) * width
    h, H = (1 - pt) * height, (1 + pt) * height

    inner = mask.all_(xs < w, ys < h)
    outer = mask.all_(xs < W, ys < H)

    return outer - inner


@ar.definition
def Triangle(p: math.Barycentric) -> mask.Mask:
    """A triangle.

    .. arlunio-image:: Triangle Demo
       :align: right

       ::

          import arlunio.shape as shape
          import arlunio.image as img

          tri = shape.Triangle()
          image = img.fill(tri(width=256, height=256))

    """

    l1 = p[:, :, 0]
    l2 = p[:, :, 1]
    l3 = p[:, :, 2]

    return mask.all_(0 < l1, l1 < 1, 0 < l2, l2 < 1, 0 < l3, l3 < 1)
